from dataclasses import dataclass
from typing import Literal

from starknet_py.net.models import StarknetChainId


@dataclass
class Network:
    chain_id: int
    gateway_url: str


# network configurations
LOCAL = Network(StarknetChainId.TESTNET, "http://127.0.0.1:5050")
TESTNET = Network(StarknetChainId.TESTNET, "https://alpha4.starknet.io")
INTEGRATION = Network(
    StarknetChainId.TESTNET, "https://external.integration.starknet.io"
)
MAINNET = Network(StarknetChainId.MAINNET, "https://alpha-mainnet.starknet.io")

NETWORKS = {
    "testnet": TESTNET,
    "integration": INTEGRATION,
    "mainnet": MAINNET,
    "local": LOCAL,
}

NetworkType = Literal[
    "testnet",
    "integration",
    "mainnet",
    "local",
]


# contract address configuration
@dataclass
class ContractAddresses:
    publisher_registry_address: int
    oracle_proxy_address: int


TESTNET_CONTRACTS = ContractAddresses(
    0x1ACCF0E75CBA1C5F3193D0DF912AB01BA330A5365E903754B30169625DD8E2A,
    0x21D3A3150FA4C2EF5C6C1EE718BB04355E3AE619719B0369A391FCB8AE36FDF,
)
INTEGRATION_CONTRACTS = ContractAddresses(
    0x062F25D256931E058906542CEE48D645996F0D84C98940AE7D2BD6B68ECC2925,
    0x072D2CD06603E0C957F54A4A1909745939FCFD597EA22323672A6EACBF27D02A,
)
MAINNET_CONTRACTS = ContractAddresses(
    0x051949605AB53FCC2C0ADC1D53A72DD0FBCBF83E52399A8B05552F675B1DB4E9,
    0x00225A37DE623DBD4D2287DDED4E0CB0EB4A5D7D9051D0E89A1321D4BCF9FDB2,
)

CONTRACT_ADDRESSES = {
    "testnet": TESTNET_CONTRACTS,
    "integration": INTEGRATION_CONTRACTS,
    "mainnet": MAINNET_CONTRACTS,
}

EnvironmentTypes = Literal[
    "testnet",
    "integration",
    "mainnet",
]



from enum import IntEnum, unique
from typing import Literal

ADDRESS = int
HEX_STR = str

# Network Types
STAGING = "staging"
TESTNET = "testnet"
INTEGRATION = "integration"
MAINNET = "mainnet"

Network = Literal["staging", "testnet", "integration", "mainnet"]

CHAIN_IDS = {
    INTEGRATION: 1536727068981429685321,
    TESTNET: 1536727068981429685321,
    MAINNET: 23448594291968334,
}

GATEWAY_URLS = {
    TESTNET: "https://alpha4.starknet.io",
    INTEGRATION: "https://external.integration.starknet.io",
    MAINNET: "https://alpha-mainnet.starknet.io",
}


# aggregation mode enum
@unique
class AggregationMode(IntEnum):
    MEDIAN = 84959893733710  # str_to_felt("MEDIAN")



import configparser
from pathlib import Path

import typer
from empiric.cli import SUCCESS
from starknet_py.net import KeyPair
from starknet_py.net.client import Client
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models.address import AddressRepresentation
from starknet_py.transactions.deploy import make_deploy_tx

from .compiled_account_contract import COMPILED_ACCOUNT_CONTRACT


async def deploy_account_contract(
    client: Client, public_key: int
) -> AddressRepresentation:
    deploy_tx = make_deploy_tx(
        constructor_calldata=[public_key],
        compiled_contract=COMPILED_ACCOUNT_CONTRACT,
    )
    result = await client.deploy(deploy_tx)
    await client.wait_for_tx(
        tx_hash=result.transaction_hash,
        wait_for_accept=True,
    )
    return result.contract_address


async def create_account(client: GatewayClient, config_file: Path):
    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)

    account_private_key = int(config_parser["SECRET"]["private-key"])

    key_pair = KeyPair.from_private_key(account_private_key)
    address = await deploy_account_contract(client, key_pair.public_key)
    typer.echo(f"created address: {address}")

    config_parser["USER"]["address"] = str(address)
    with open(config_file, "w") as f:
        config_parser.write(f)

    return SUCCESS



from .bitstamp import BitstampFetcher
from .cex import CexFetcher
from .coinbase import CoinbaseFetcher
from .coingecko import CoingeckoFetcher
from .ftx import FtxFetcher
from .gemini import GeminiFetcher
from .thegraph import TheGraphFetcher



import datetime
import hmac
import logging
import os
import re
import time
from typing import Dict, List

import requests
from aiohttp import ClientSession
from empiric.core.entry import FutureEntry, SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset
from empiric.publisher.types import PublisherInterfaceT

logger = logging.getLogger(__name__)


class FtxFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://ftx.com/api"
    SOURCE: str = "ftx"
    endpoint: str = "/markets"

    FTX_API_KEY: str
    FTX_API_SECRET: str

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

        self.FTX_API_KEY = os.environ.get("FTX_API_KEY")
        self.FTX_API_SECRET = os.environ.get("FTX_API_SECRET")

    async def fetch(self, session: ClientSession) -> List[SpotEntry]:
        market_endpoint = "/markets"
        headers = self.generate_ftx_headers(market_endpoint)

        async with session.get(
            self.BASE_URL + market_endpoint, headers=headers
        ) as resp:
            response_json = await resp.json()
            spot_data = response_json["result"]

        futures_endpoint = "/futures"
        headers = self.generate_ftx_headers(futures_endpoint)
        async with session.get(
            self.BASE_URL + futures_endpoint, headers=headers
        ) as resp:
            response_json = await resp.json()
            future_data = response_json["result"]

        return self._handle_assets(spot_data, future_data)

    def fetch_sync(self) -> List[SpotEntry]:
        market_endpoint = "/markets"
        headers = self.generate_ftx_headers(market_endpoint)

        resp = requests.get(self.BASE_URL + market_endpoint, headers=headers)
        response_json = resp.json()
        spot_data = response_json["result"]

        futures_endpoint = "/futures"
        headers = self.generate_ftx_headers(futures_endpoint)
        resp = requests.get(self.BASE_URL + futures_endpoint, headers=headers)
        response_json = resp.json()
        future_data = response_json["result"]

        return self._handle_assets(spot_data, future_data)

    def generate_ftx_headers(self, endpoint: str) -> Dict[str, str]:
        timestamp = int(time.time() * 1000)
        signature = hmac.new(
            self.FTX_API_SECRET.encode(),
            (str(timestamp) + "GET" + endpoint).encode("ascii"),
            "sha256",
        ).hexdigest()
        headers = {
            "FTX-KEY": self.FTX_API_KEY,
            "FTX-SIGN": signature,
            "FTX-TS": str(timestamp),
        }
        return headers

    def parse_ftx_spot(self, asset, data, source, publisher, timestamp) -> SpotEntry:
        pair = asset["pair"]
        pair_id = currency_pair_to_pair_id(*pair)

        result = [e for e in data if e["name"] == "/".join(pair)]
        if len(result) == 0:
            logger.debug(f"No entry found for {'/'.join(pair)} from FTX")
            return

        if len(result) != 1:
            raise ValueError(
                f"Found more than one matching entries for FTX response and price pair {pair}"
            )

        price = float(result[0]["price"])
        price_int = int(price * (10 ** asset["decimals"]))

        logger.info(f"Fetched price {price} for {'/'.join(pair)} from FTX")

        return SpotEntry(
            pair_id=pair_id,
            price=price_int,
            timestamp=timestamp,
            source=source,
            publisher=publisher,
        )

    def parse_ftx_futures(
        self, asset, data, source, publisher, timestamp
    ) -> List[FutureEntry]:
        pair = asset["pair"]
        pair_id = currency_pair_to_pair_id(*pair)
        if pair[1] != "USD":
            logger.debug(
                f"Unable to fetch price from FTX for non-USD derivative {pair}"
            )
            return

        result = [e for e in data if re.match(rf"{pair[0]}-[0-9]+", e["name"])]
        if len(result) == 0:
            logger.debug(f"No entry found for {'/'.join(pair)} from FTX")
            return

        entries = []

        for future in result:
            price = float(future["mark"])
            price_int = int(price * (10 ** asset["decimals"]))

            expiry_timestamp = int(
                datetime.datetime.strptime(
                    future["expiry"],
                    "%Y-%m-%dT%H:%M:%S%z",
                ).timestamp()
            )

            logger.info(
                f"Fetched futures price {price} for {pair_id}-{expiry_timestamp} from FTX"
            )

            entries.append(
                FutureEntry(
                    pair_id=pair_id,
                    price=price_int,
                    timestamp=timestamp,
                    source=source,
                    publisher=publisher,
                    expiry_timestamp=expiry_timestamp,
                )
            )

        return entries

    def _handle_assets(self, spot_data, future_data):
        entries = []

        timestamp = int(time.time())
        for asset in self.assets:
            if asset["type"] == "SPOT":
                entry = self.parse_ftx_spot(
                    asset, spot_data, self.SOURCE, self.publisher, timestamp
                )
                if entry is not None:
                    entries.append(entry)
                continue
            elif asset["type"] == "FUTURE":
                future_entries = self.parse_ftx_futures(
                    asset, future_data, self.SOURCE, self.publisher, timestamp
                )
                if len(future_entries) > 0:
                    entries.extend(future_entries)
                continue
            else:
                logger.debug(f"Unable to fetch FTX for un-supported asset type {asset}")

        return entries



import asyncio
import base64
import datetime
import hmac
import logging
import os
from hashlib import sha256
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class CoinbaseFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://api.exchange.coinbase.com"
    SOURCE: str = "coinbase"
    REQUEST_PATH = "/oracle"
    METHOD = "GET"

    COINBASE_API_SECRET: str
    COINBASE_API_KEY: str
    COINBASE_API_PASSPHRASE: str
    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        # TODO (rlkelly): migrate from using environment variables
        self.COINBASE_API_SECRET = os.environ.get("COINBASE_API_SECRET")
        self.COINBASE_API_KEY = os.environ.get("COINBASE_API_KEY")
        self.COINBASE_API_PASSPHRASE = os.environ.get("COINBASE_API_PASSPHRASE")

        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        if pair[1] != "USD":
            logger.debug(
                f"Unable to fetch Coinbase price for non-USD denomination {pair[1]}"
            )
            return PublisherFetchError(f"NON-USD PAIR: {pair}")

        request_timestamp = str(
            int(
                datetime.datetime.now(datetime.timezone.utc)
                .replace(tzinfo=datetime.timezone.utc)
                .timestamp()
            )
        )

        signature = hmac.new(
            base64.b64decode(self.COINBASE_API_SECRET),
            (request_timestamp + self.METHOD + self.REQUEST_PATH).encode("ascii"),
            sha256,
        )

        headers = {
            "Accept": "application/json",
            "CB-ACCESS-KEY": self.COINBASE_API_KEY,
            "CB-ACCESS-SIGN": base64.b64encode(signature.digest()).decode("utf8"),
            "CB-ACCESS-TIMESTAMP": request_timestamp,
            "CB-ACCESS-PASSPHRASE": self.COINBASE_API_PASSPHRASE,
        }

        async with session.get(
            self.BASE_URL + self.REQUEST_PATH, headers=headers
        ) as resp:
            result = await resp.json()
            return self._construct(asset, result)

    def _fetch_pair_sync(
        self, asset: EmpiricSpotAsset
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        if pair[1] != "USD":
            logger.debug(
                f"Unable to fetch Coinbase price for non-USD denomination {pair[1]}"
            )
            return PublisherFetchError(f"NON-USD PAIR: {pair}")

        request_timestamp = str(
            int(
                datetime.datetime.now(datetime.timezone.utc)
                .replace(tzinfo=datetime.timezone.utc)
                .timestamp()
            )
        )

        signature = hmac.new(
            base64.b64decode(self.COINBASE_API_SECRET),
            (request_timestamp + self.METHOD + self.REQUEST_PATH).encode("ascii"),
            sha256,
        )

        headers = {
            "Accept": "application/json",
            "CB-ACCESS-KEY": self.COINBASE_API_KEY,
            "CB-ACCESS-SIGN": base64.b64encode(signature.digest()),
            "CB-ACCESS-TIMESTAMP": request_timestamp,
            "CB-ACCESS-PASSPHRASE": self.COINBASE_API_PASSPHRASE,
        }

        resp = requests.get(self.BASE_URL + self.REQUEST_PATH, headers=headers)
        result = resp.json()
        return self._construct(asset, result)

    async def fetch(
        self, session: ClientSession
    ) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Coinbase for non-spot asset {asset}")
                continue

            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries)

    def fetch_sync(self) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Coinbase for non-spot asset {asset}")
                continue

            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        pair_id = currency_pair_to_pair_id(*pair)

        if pair[0] in result["prices"]:
            price = float(result["prices"][pair[0]])
            price_int = int(price * (10 ** asset["decimals"]))

            timestamp = int(result["timestamp"])

            logging.info(f"Fetched price {price} for {pair_id} from Coinbase")

            return SpotEntry(
                pair_id=pair_id,
                price=price_int,
                timestamp=timestamp,
                source=self.SOURCE,
                publisher=self.publisher,
            )

        return PublisherFetchError(f"No entry found for {pair_id} from Coinbase")



import collections
import logging
from typing import List

from empiric.core.contract import Contract
from empiric.core.entry import SpotEntry
from empiric.core.types import AggregationMode
from empiric.core.utils import str_to_felt
from starknet_py.contract import InvokeResult
from starknet_py.net.client import Client

logger = logging.getLogger(__name__)

OracleResponse = collections.namedtuple(
    "OracleResponse",
    ["price", "decimals", "last_updated_timestamp", "num_sources_aggregated"],
)


class OracleMixin:
    publisher_registry: Contract
    client: Client

    async def publish_spot_entry(
        self,
        pair_id: int,
        value: int,
        timestamp: int,
        source: int,
        publisher: int,
        volume: int = 0,
        max_fee: int = int(1e18),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.oracle.publish_spot_entry.invoke(
            {
                "pair_id": pair_id,
                "value": value,
                "timestamp": timestamp,
                "source": source,
                "publisher": publisher,
            },
            max_fee=max_fee,
        )
        return invocation

    async def publish_many(
        self, entries: List[SpotEntry], pagination=0, max_fee=int(1e18)
    ) -> List[InvokeResult]:
        if len(entries) == 0:
            logger.warn("Skipping publishing as entries array is empty")
            return

        invocations = []
        # TODO (this PR) filter by entry.type and publish Spot, Future and Generic entries separately
        if pagination:
            ix = 0
            while ix < len(entries):
                invocation = await self.oracle.publish_spot_entries.invoke(
                    SpotEntry.serialize_entries(entries[ix : ix + pagination]),
                    callback=self.track_nonce,
                    max_fee=max_fee,
                )
                logger.info(str(invocation))
                ix += pagination
                invocations.append(invocation)
        else:
            invocation = await self.oracle.publish_spot_entries.invoke(
                SpotEntry.serialize_entries(entries), max_fee=max_fee
            )
            invocations.append(invocation)
            logger.info(str(invocation))

        logger.info(
            f"Sent {len(entries)} updated entries with transaction {invocation.hash}"
        )

        return invocations

    async def get_entries(self, pair_id, sources=[]) -> List[SpotEntry]:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )
        response = await self.oracle.get_spot_entries.call(pair_id, sources)

        return [SpotEntry.from_dict(entry) for entry in response.entries]

    async def get_spot(
        self,
        pair_id,
        aggregation_mode: AggregationMode = AggregationMode.MEDIAN,
        sources=None,
    ) -> OracleResponse:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )
        if sources is None:
            response = await self.oracle.get_spot.call(
                pair_id,
                aggregation_mode.value,
            )
        else:
            response = await self.oracle.get_spot_entries_for_sources.call(
                pair_id, aggregation_mode.value, sources
            )

        return OracleResponse(
            response.price,
            response.decimals,
            response.last_updated_timestamp,
            response.num_sources_aggregated,
        )

    async def get_future(
        self,
        pair_id,
        expiry_timestamp,
        aggregation_mode: AggregationMode = AggregationMode.MEDIAN,
        # TODO Add sources on the oracle contract and then in the client here
        # sources=None,
    ) -> OracleResponse:
        if isinstance(pair_id, str):
            pair_id = str_to_felt(pair_id)
        elif not isinstance(pair_id, int):
            raise TypeError(
                "Pair ID must be string (will be converted to felt) or integer"
            )

        response = await self.oracle.get_futures.call(
            pair_id,
            expiry_timestamp,
            aggregation_mode.value,
        )

        return OracleResponse(
            response.price,
            response.decimals,
            response.last_updated_timestamp,
            response.num_sources_aggregated,
        )

    async def set_checkpoint(
        self,
        pair_id: int,
        aggregation_mode: int = str_to_felt("MEDIAN"),
        max_fee=int(1e16),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.oracle.set_checkpoint.invoke(
            pair_id,
            aggregation_mode,
            callback=self.track_nonce,
            max_fee=max_fee,
        )
        return invocation

    async def set_checkpoints(
        self,
        pair_ids: List[int],
        aggregation_mode: int = str_to_felt("MEDIAN"),
        max_fee=int(1e18),
    ) -> InvokeResult:
        if not self.is_user_client:
            raise AttributeError(
                "Must set account.  You may do this by invoking self._setup_account_client(private_key, account_contract_address)"
            )
        invocation = await self.oracle.set_checkpoints.invoke(
            pair_ids,
            aggregation_mode,
            callback=self.track_nonce,
            max_fee=max_fee,
        )
        return invocation
